"""Convenience functions built on top of `make_vjp`."""
from __future__ import absolute_import
import autograd.numpy as np
from autograd.core import make_vjp, vspace, primitive, unbox_if_possible
from autograd.container_types import make_tuple
from .errors import add_error_hints
from collections import OrderedDict
from inspect import getargspec
import warnings

def grad(fun, argnum=0):
    """
    Returns a function which computes the gradient of `fun` with respect to
    positional argument number `argnum`. The returned function takes the same
    arguments as `fun`, but returns the gradient instead. The function `fun`
    should be scalar-valued. The gradient has the same type as the argument."""
    @attach_name_and_doc(fun, argnum, 'Gradient')
    @add_error_hints
    def gradfun(*args,**kwargs):
        args = list(args)
        args[argnum] = safe_type(args[argnum])
        vjp, ans = make_vjp(fun, argnum)(*args, **kwargs)
        return vjp(vspace(ans).ones())

    return gradfun

def jacobian(fun, argnum=0):
    """
    Returns a function which computes the Jacobian of `fun` with respect to
    positional argument number `argnum`, which must be a scalar or array. Unlike
    `grad` it is not restricted to scalar-output functions, but also it cannot
    take derivatives with respect to some argument types (like lists or dicts).
    If the input to `fun` has shape (in1, in2, ...) and the output has shape
    (out1, out2, ...) then the Jacobian has shape (out1, out2, ..., in1, in2, ...).
    """
    @attach_name_and_doc(fun, argnum, 'Jacobian')
    @add_error_hints
    def jacfun(*args, **kwargs):
        vjp, ans = make_vjp(fun, argnum)(*args, **kwargs)
        ans_vspace = vspace(ans)
        jacobian_shape = ans_vspace.shape + vspace(args[argnum]).shape
        grads = map(vjp, ans_vspace.standard_basis())
        return np.reshape(np.stack(grads), jacobian_shape)

    return jacfun

def grad_named(fun, argname):
    '''Takes gradients with respect to a named argument.
       Doesn't work on *args or **kwargs.'''
    arg_index = getargspec(fun).args.index(argname)
    return grad(fun, arg_index)

def value_and_multigrad(fun, argnums=[0]):
    """Returns a function that returns both value and gradients wrt multiple
    arguments simultaneously."""
    def combined_arg_fun(multi_arg, *args, **kwargs):
        extra_args_list = list(args)
        for argnum_ix, arg_ix in enumerate(argnums):
            extra_args_list[arg_ix] = multi_arg[argnum_ix]
        return fun(*extra_args_list, **kwargs)
    gradfun = value_and_grad(combined_arg_fun, argnum=0)
    def gradfun_rearranged(*args, **kwargs):
        multi_arg = make_tuple(*[args[i] for i in argnums])
        return gradfun(multi_arg, *args, **kwargs)
    return gradfun_rearranged

def multigrad(fun, argnums=[0]):
    """Returns a function that returns gradients wrt multiple arguments
    simultaneously."""
    double_val_fun = value_and_multigrad(fun, argnums=argnums)
    def multigrad_fun(*args, **kwargs):
        return double_val_fun(*args, **kwargs)[1]
    return multigrad_fun

elementwise_grad = grad  # backward compatibility

def hessian(fun, argnum=0):
    "Returns a function that computes the exact Hessian."
    return jacobian(jacobian(fun, argnum), argnum)

def make_hvp(fun, argnum=0):
    """Builds a function for evaluating the Hessian-vector product at a point,
    which may be useful when evaluating many Hessian-vector products at the same
    point while caching the results of the forward pass."""
    def hvp_maker(*args, **kwargs):
        return make_vjp(grad(fun, argnum), argnum)(*args, **kwargs)
    return hvp_maker

def hessian_tensor_product(fun, argnum=0):
    """Builds a function that returns the exact Hessian-tensor product.
    The returned function has arguments (*args, tensor, **kwargs), and for
    vectors takes roughly 4x as long to evaluate as the original function."""
    fun_grad = grad(fun, argnum)
    def vector_dot_grad(*args, **kwargs):
        args, vector = args[:-1], args[-1]
        return np.tensordot(fun_grad(*args, **kwargs), vector, np.ndim(vector))
    return grad(vector_dot_grad, argnum)
hessian_vector_product = hessian_tensor_product

def tensor_jacobian_product(fun, argnum=0):
    """Builds a function that returns the exact tensor-Jacobian product, that
    is the Jacobian matrix left-multiplied by tensor. The returned function
    has arguments (*args, tensor, **kwargs)."""
    def vector_dot_fun(*args, **kwargs):
        args, vector = args[:-1], args[-1]
        return np.tensordot(vector, fun(*args, **kwargs), axes=np.ndim(vector))
    return jacobian(vector_dot_fun, argnum)
vector_jacobian_product = tensor_jacobian_product

def make_jvp(fun, argnum=0):
    """Builds a function for evaluating the Jacobian-vector product at a
    point. Roughly 1.5x more FLOPs than forward-mode, plus memory requirements
    that scale with the number of primitives applied in the evaluation of f, as
    well as other overheads. See j-towns.github.io/2017/06/12/A-new-trick.html
    and github.com/BB-UCL/autograd-forward."""
    def jvp_maker(*args, **kwargs):
        vjp, y = make_vjp(fun, argnum)(*args, **kwargs)
        vjp_vjp, _ = make_vjp(vjp)(vspace(y).zeros())
        return vjp_vjp  # vjp_vjp is just jvp by linearity
    return jvp_maker

def make_ggnvp(f, g=lambda x: 1./2*np.sum(x**2, axis=-1), f_argnum=0):
    """Builds a function for evaluating generalized-Gauss-Newton-vector products
    at a point. Slightly more expensive than mixed-mode."""
    def ggnvp_maker(*args, **kwargs):
        f_vjp, f_x = make_vjp(f, f_argnum)(*args, **kwargs)
        g_hvp, grad_g_x = make_vjp(grad(g))(f_x)
        f_jvp, _ = make_vjp(f_vjp)(vspace(grad_g_x).zeros())
        def ggnvp(v): return f_vjp(g_hvp(f_jvp(v)))
        return ggnvp
    return ggnvp_maker

def value_and_grad(fun, argnum=0):
    """Returns a function that returns both value and gradient. Suitable for use
    in scipy.optimize"""
    def double_val_fun(*args, **kwargs):
        val = fun(*args, **kwargs)
        return make_tuple(val, unbox_if_possible(val))
    gradval_and_val = grad_and_aux(double_val_fun, argnum)
    flip = lambda x, y: make_tuple(y, x)
    return lambda *args, **kwargs: flip(*gradval_and_val(*args, **kwargs))

def grad_and_aux(fun, argnum=0):
    """Builds a function that returns the gradient of the first output and the
    (unmodified) second output of a function that returns two outputs."""
    def grad_and_aux_fun(*args, **kwargs):
        saved = lambda: None
        def return_val_save_aux(*args, **kwargs):
            val, saved.aux = fun(*args, **kwargs)
            return val
        gradval = grad(return_val_save_aux, argnum)(*args, **kwargs)
        return gradval, saved.aux

    return grad_and_aux_fun

def multigrad_dict(fun):
    "Takes gradients wrt all arguments simultaneously,"
    "returns a dict mapping 'argname' to 'gradval'"

    import funcsigs
    sig = funcsigs.signature(fun)

    def select(preds, lst):
        idx = lambda item: next(
            (i for i, pred in enumerate(preds) if pred(item)), len(preds))
        results = [[] for _ in preds] + [[]]
        for item in lst:
            results[idx(item)].append(item)
        return results

    is_var_pos = lambda name: sig.parameters[name].kind == sig.parameters[name].VAR_POSITIONAL
    is_var_kwd = lambda name: sig.parameters[name].kind == sig.parameters[name].VAR_KEYWORD
    var_pos, var_kwd, argnames = select([is_var_pos, is_var_kwd], sig.parameters)

    todict = lambda dct: {key:dct[key] for key in dct}

    def apply_defaults(arguments):
        defaults = {name: param.default for name, param in sig.parameters.items()
                    if param.default is not param.empty}
        return OrderedDict((name, arguments[name] if name in arguments else defaults[name])
                           for name in sig.parameters)

    def gradfun(*args, **kwargs):
        bindings = sig.bind(*args, **kwargs)

        args = lambda dct: tuple(dct[var_pos[0]]) if var_pos else ()
        kwargs = lambda dct: todict(dct[var_kwd[0]]) if var_kwd else {}
        others = lambda dct: tuple(dct[argname] for argname in argnames
                                   if argname not in var_kwd + var_pos)

        newfun = lambda dct: fun(*(others(dct) + args(dct)), **kwargs(dct))

        argdict = apply_defaults(bindings.arguments)
        grad_dict = grad(newfun)(dict(argdict))
        return OrderedDict((argname, grad_dict[argname]) for argname in argdict)

    return gradfun

def checkpoint(fun):
    """Returns a checkpointed version of `fun`, where intermediate values
    computed during the forward pass of `fun` are discarded and then recomputed
    for the backward pass. Useful to save memory, effectively trading off time
    and memory. See e.g. arxiv.org/abs/1604.06174.
    """
    def wrapped_grad(argnum, g, ans, vs, gvs, args, kwargs):
        return make_vjp(fun, argnum)(*args, **kwargs)[0](g)
    wrapped = primitive(fun)
    wrapped.vjp = wrapped_grad
    return wrapped

def attach_name_and_doc(fun, argnum, opname):
    namestr = "{op}_{fun}_wrt_argnum_{argnum}".format(
        op=opname.lower(), fun=getattr(fun, '__name__', '[unknown name]'), argnum=argnum)
    docstr = "{op} of function {fun} with respect to argument number {argnum}. " \
        "Has the same arguments as {fun} but the return value has type of " \
        "argument {argnum}.".format(op=opname, fun=getattr(fun, '__name__', '[unknown name]'),
        argnum=argnum)

    def wrap(gradfun):
        try:
            gradfun.__name__ = namestr
            gradfun.__doc__ = docstr
        finally:
            return gradfun
    return wrap

def safe_type(value):
    if isinstance(value, int):
        warnings.warn("Casting int to float to handle differentiation.")
        return float(value)
    else:
        return value

def cast_to_same_dtype(value, example):
    if hasattr(example, 'dtype') and example.dtype.type is not np.float64:
        return np.array(value, dtype=example.dtype)
    else:
        return value



from __future__ import absolute_import
import numpy as onp
from numpy.core.einsumfunc import _parse_einsum_input

from autograd.core import primitive, getval, vspace
from . import numpy_wrapper as anp
from .numpy_extra import ArrayNode, take
from builtins import range, zip
from future.utils import string_types

# ----- Functions that are constant w.r.t. continuous inputs -----

anp.where.defvjp_is_zero(argnums=(0,))
anp.nan_to_num.defvjp(lambda g, ans, vs, gvs, x: anp.where(anp.isfinite(x), g, 0.))

# ----- Binary ufuncs -----

anp.add.defvjp(        lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g))
anp.add.defvjp(        lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g), argnum=1)
anp.multiply.defvjp(   lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, y * g))
anp.multiply.defvjp(   lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, x * g), argnum=1)
anp.subtract.defvjp(   lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g))
anp.subtract.defvjp(   lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, -g), argnum=1)
anp.divide.defvjp(     lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs,   g / y))
anp.divide.defvjp(     lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, - g * x / y**2), argnum=1)
anp.maximum.defvjp(    lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(x, ans, y)))
anp.maximum.defvjp(    lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(y, ans, x)), argnum=1)
anp.minimum.defvjp(    lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(x, ans, y)))
anp.minimum.defvjp(    lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(y, ans, x)), argnum=1)
anp.fmax.defvjp(       lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(x, ans, y)))
anp.fmax.defvjp(       lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(y, ans, x)), argnum=1)
anp.fmin.defvjp(       lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(x, ans, y)))
anp.fmin.defvjp(       lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * balanced_eq(y, ans, x)), argnum=1)
anp.logaddexp.defvjp(  lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * anp.exp(x-ans)))
anp.logaddexp.defvjp(  lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * anp.exp(y-ans)), argnum=1)
anp.logaddexp2.defvjp( lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * 2**(x-ans)))
anp.logaddexp2.defvjp( lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g * 2**(y-ans)), argnum=1)
anp.true_divide.defvjp(lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g / y))
anp.true_divide.defvjp(lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, - g * x / y**2), argnum=1)
anp.mod.defvjp(        lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g))
anp.remainder.defvjp(  lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, g))
anp.mod.defvjp(        lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, -g * anp.floor(x/y)), argnum=1)
anp.remainder.defvjp(  lambda g, ans, vs, gvs, x, y : unbroadcast(vs, gvs, -g * anp.floor(x/y)), argnum=1)
anp.power.defvjp(
    lambda g, ans, vs, gvs, x, y :
    unbroadcast(vs, gvs, g * y * x ** anp.where(y, y - 1, 1.)))
anp.power.defvjp(
    lambda g, ans, vs, gvs, x, y :
    unbroadcast(vs, gvs, g * anp.log(replace_zero(x, 1.)) * x ** y), argnum=1)


# ----- Simple grads -----

anp.negative.defvjp(lambda g, ans, vs, gvs, x: -g)
anp.abs.defvjp(
    lambda g, ans, vs, gvs, x : g * replace_zero(anp.conj(x), 0.) / replace_zero(ans, 1.))
anp.fabs.defvjp(    lambda g, ans, vs, gvs, x : anp.sign(x) * g)  # fabs doesn't take complex numbers.
anp.absolute.defvjp(lambda g, ans, vs, gvs, x : g * anp.conj(x) / ans)
anp.reciprocal.defvjp(lambda g, ans, vs, gvs, x : - g / x**2)
anp.exp.defvjp(   lambda g, ans, vs, gvs, x : ans * g)
anp.exp2.defvjp(  lambda g, ans, vs, gvs, x : ans * anp.log(2) * g)
anp.expm1.defvjp( lambda g, ans, vs, gvs, x : (ans + 1) * g)
anp.log.defvjp(   lambda g, ans, vs, gvs, x : g / x)
anp.log2.defvjp(  lambda g, ans, vs, gvs, x : g / x / anp.log(2))
anp.log10.defvjp( lambda g, ans, vs, gvs, x : g / x / anp.log(10))
anp.log1p.defvjp( lambda g, ans, vs, gvs, x : g / (x + 1))
anp.sin.defvjp(   lambda g, ans, vs, gvs, x : g * anp.cos(x))
anp.cos.defvjp(   lambda g, ans, vs, gvs, x : - g * anp.sin(x))
anp.tan.defvjp(   lambda g, ans, vs, gvs, x : g / anp.cos(x) **2)
anp.arcsin.defvjp(lambda g, ans, vs, gvs, x : g / anp.sqrt(1 - x**2))
anp.arccos.defvjp(lambda g, ans, vs, gvs, x :-g / anp.sqrt(1 - x**2))
anp.arctan.defvjp(lambda g, ans, vs, gvs, x : g / (1 + x**2))
anp.sinh.defvjp(  lambda g, ans, vs, gvs, x : g * anp.cosh(x))
anp.cosh.defvjp(  lambda g, ans, vs, gvs, x : g * anp.sinh(x))
anp.tanh.defvjp(  lambda g, ans, vs, gvs, x : g / anp.cosh(x) **2)
anp.arcsinh.defvjp(lambda g, ans, vs, gvs, x : g / anp.sqrt(x**2 + 1))
anp.arccosh.defvjp(lambda g, ans, vs, gvs, x : g / anp.sqrt(x**2 - 1))
anp.arctanh.defvjp(lambda g, ans, vs, gvs, x : g / (1 - x**2))
anp.rad2deg.defvjp(lambda g, ans, vs, gvs, x : g / anp.pi * 180.0)
anp.degrees.defvjp(lambda g, ans, vs, gvs, x : g / anp.pi * 180.0)
anp.deg2rad.defvjp(lambda g, ans, vs, gvs, x : g * anp.pi / 180.0)
anp.radians.defvjp(lambda g, ans, vs, gvs, x : g * anp.pi / 180.0)
anp.square.defvjp( lambda g, ans, vs, gvs, x : g * 2 * x)
anp.sqrt.defvjp(   lambda g, ans, vs, gvs, x : g * 0.5 * x**-0.5)
anp.sinc.defvjp(   lambda g, ans, vs, gvs, x : g * (anp.cos(anp.pi*x)*anp.pi*x - anp.sin(anp.pi*x))/(anp.pi*x**2))
anp.reshape.defvjp(lambda g, ans, vs, gvs, x, shape, order=None : anp.reshape(g, vs.shape, order=order))
anp.roll.defvjp(   lambda g, ans, vs, gvs, x, shift, axis=None  : anp.roll(g, -shift, axis=axis))
anp.array_split.defvjp(lambda g, ans, vs, gvs, ary, idxs, axis=0 : anp.concatenate(g, axis=axis))
anp.split.defvjp(      lambda g, ans, vs, gvs, ary, idxs, axis=0 : anp.concatenate(g, axis=axis))
anp.vsplit.defvjp(     lambda g, ans, vs, gvs, ary, idxs         : anp.concatenate(g, axis=0))
anp.hsplit.defvjp(     lambda g, ans, vs, gvs, ary, idxs         : anp.concatenate(g, axis=1))
anp.dsplit.defvjp(     lambda g, ans, vs, gvs, ary, idxs         : anp.concatenate(g, axis=2))
anp.ravel.defvjp(  lambda g, ans, vs, gvs, x, order=None   : anp.reshape(g, vs.shape, order=order))
anp.expand_dims.defvjp(lambda g, ans, vs, gvs, x, axis     : anp.reshape(g, vs.shape))
anp.squeeze.defvjp(lambda g, ans, vs, gvs, x, axis=None    : anp.reshape(g, vs.shape))
anp.diag.defvjp(   lambda g, ans, vs, gvs, x, k=0          : anp.diag(g, k))
anp.flipud.defvjp( lambda g, ans, vs, gvs, x,              : anp.flipud(g))
anp.fliplr.defvjp( lambda g, ans, vs, gvs, x,              : anp.fliplr(g))
anp.rot90.defvjp(  lambda g, ans, vs, gvs, x, k=1          : anp.rot90(g, -k))
anp.trace.defvjp(  lambda g, ans, vs, gvs, x, offset=0     :
                    anp.einsum('ij,...->ij...', anp.eye(x.shape[0], x.shape[1], k=offset), g))
anp.full.defvjp(   lambda g, ans, vs, gvs, shape, fill_value, dtype=None : anp.sum(g), argnum=1)
anp.triu.defvjp(   lambda g, ans, vs, gvs, x, k=0          : anp.triu(g, k=k))
anp.tril.defvjp(   lambda g, ans, vs, gvs, x, k=0          : anp.tril(g, k=k))
anp.clip.defvjp(   lambda g, ans, vs, gvs, x, a_min, a_max : g * anp.logical_and(ans != a_min, ans != a_max))
anp.swapaxes.defvjp(lambda g, ans, vs, gvs, x, axis1, axis2: anp.swapaxes(g, axis2, axis1))
anp.moveaxis.defvjp(lambda g, ans, vs, gvs, a, source, destination:
                    anp.moveaxis(g, destination, source))
anp.rollaxis.defvjp(lambda g, ans, vs, gvs, a, axis, start=0: anp.rollaxis(g, start - 1, axis) if start > axis
                                                 else anp.rollaxis(g, start, axis + 1))
anp.real_if_close.defvjp(lambda g, ans, vs, gvs, x : match_complex(vs, g))
anp.real.defvjp(  lambda g, ans, vs, gvs, x   : match_complex(vs, g))
anp.imag.defvjp(  lambda g, ans, vs, gvs, x   : match_complex(vs, -1j * g))
anp.conj.defvjp(  lambda g, ans, vs, gvs, x   : anp.conj(g))
anp.conjugate.defvjp(lambda g, ans, vs, gvs, x: anp.conj(g))
anp.angle.defvjp( lambda g, ans, vs, gvs, x   : match_complex(vs, g * anp.conj(x * 1j) / anp.abs(x)**2))
anp.where.defvjp( lambda g, ans, vs, gvs, c, x=None, y=None : anp.where(c, g, anp.zeros(g.shape)), argnum=1)
anp.where.defvjp( lambda g, ans, vs, gvs, c, x=None, y=None : anp.where(c, anp.zeros(g.shape), g), argnum=2)
anp.cross.defvjp(lambda g, ans, vs, gvs, a, b, axisa=-1, axisb=-1, axisc=-1, axis=None :
                  anp.cross(b, g, axisb, axisc, axisa, axis), argnum=0)
anp.cross.defvjp(lambda g, ans, vs, gvs, a, b, axisa=-1, axisb=-1, axisc=-1, axis=None :
                  anp.cross(g, a, axisc, axisa, axisb, axis), argnum=1)
anp.linspace.defvjp(lambda g, ans, vs, gvs, start, stop, num : anp.dot(anp.linspace(1.0, 0.0, num), g))
anp.linspace.defvjp(lambda g, ans, vs, gvs, start, stop, num : anp.dot(anp.linspace(0.0, 1.0, num), g), argnum=1)

# ----- Trickier grads -----

def grad_diff(g, ans, vs, gvs, a, n=1, axis=-1):
    nd = len(vs.shape)
    sl1 = [slice(None)]*nd
    sl1[axis] = slice(None, 1)

    sl2 = [slice(None)]*nd
    sl2[axis] = slice(-1, None)

    def undiff(g):
        if g.shape[axis] > 0:
            return anp.concatenate((-g[sl1], -anp.diff(g, axis=axis), g[sl2]), axis=axis)
        shape = list(gvs.shape)
        shape[axis] = 1
        return anp.zeros(shape)

    def helper(g, n):
        if n == 0:
            return g
        return helper(undiff(g), n-1)
    return helper(g, n)

anp.diff.defvjp(grad_diff)

def grad_repeat(g, ans, vs, gvs, x, repeats, axis=None):
    shape = x.shape
    if axis is None:  # If axis is none, np.repeat() repeats the flattened array.
        expanded = anp.reshape(g, (anp.prod(shape),) + (repeats,))
        return anp.reshape(anp.sum(expanded, axis=1, keepdims=False), shape)
    else:
        if shape[axis] == 1:  # For this common case, the logic is simple.
            return anp.sum(g, axis=axis, keepdims=True)
        else:
            expanded = anp.reshape(g, shape[0:axis+1] + (repeats,) + shape[axis+1:])
            return anp.sum(expanded, axis=axis+1, keepdims=False)
anp.repeat.defvjp(grad_repeat)

def grad_tile(g, ans, vs, gvs, x, reps):
    reps = [reps] if anp.isscalar(reps) else reps
    for axis, rep in enumerate(reps):
        g = sum(anp.split(g, rep, axis))
    return anp.reshape(g, x.shape)
anp.tile.defvjp(grad_tile)

def grad_kron(argnum, G, ans, vs, gvs, orig_A, orig_B):
    # kron has different promotion rules than dot. the reshapes are necessary if
    # and only if (1) orig_B is 1D or (2) orig_A and/or orig_B are 0D
    A, B = anp.atleast_2d(orig_A), anp.atleast_2d(orig_B)
    shape = list(A.shape + B.shape)
    n = anp.ndim(A)
    shape[n-1], shape[n] = shape[n], shape[n-1]
    reshaped_G = anp.swapaxes(anp.reshape(G, shape), n-1, n)
    if argnum == 0:
        return anp.reshape(anp.tensordot(reshaped_G, B, axes=anp.ndim(B)), vs.shape)
    else:
        return anp.reshape(anp.tensordot(A, reshaped_G, axes=anp.ndim(A)), vs.shape)
anp.kron.defvjps(grad_kron, [0, 1])

def grad_transpose(g, ans, vs, gvs, x, axes=None):
    if axes is not None:
        axes = anp.argsort(axes)
    return anp.transpose(g, axes)
anp.transpose.defvjp(grad_transpose)

def repeat_to_match_shape(g, vs, axis, keepdims):
    """Returns the array g repeated along axis to fit vector space vs.
       Also returns the number of repetitions of the array."""
    if vs.shape == ():
      return g, 1
    axis = list(axis) if isinstance(axis, tuple) else axis
    shape = onp.array(vs.shape)
    shape[axis] = 1
    num_reps = onp.prod(onp.array(vs.shape)[axis])
    return anp.reshape(g, shape) + vs.zeros(), num_reps

def grad_np_sum(g, ans, vs, gvs, x, axis=None, keepdims=False, dtype=None):
    return repeat_to_match_shape(g, vs, axis, keepdims)[0]
anp.sum.defvjp(grad_np_sum)

def grad_np_mean(g, ans, vs, gvs, x, axis=None, keepdims=False):
    g_repeated, num_reps = repeat_to_match_shape(g, vs, axis, keepdims)
    return g_repeated / num_reps
anp.mean.defvjp(grad_np_mean)

def grad_np_prod(g, ans, vs, gvs, x, axis=None, keepdims=False): # TODO: Support tuples of axes.
    g_repeated, _ = repeat_to_match_shape(g * ans, vs, axis, keepdims)
    return g_repeated / x
anp.prod.defvjp(grad_np_prod)

def grad_np_var(g, ans, vs, gvs, x, axis=None, ddof=0, keepdims=False):
    if vs.iscomplex:
        g = g + 0j
    g_repeated, num_reps = repeat_to_match_shape(g, vs, axis, keepdims)
    x_minus_mean = anp.conj(x - anp.mean(x, axis=axis, keepdims=True))
    return 2.0 * g_repeated * x_minus_mean / (num_reps - ddof)
anp.var.defvjp(grad_np_var)

def grad_np_std(g, ans, vs, gvs, x, axis=None, ddof=0, keepdims=False):
    if vs.iscomplex:
        g = g + 0j
    g_repeated, num_reps = repeat_to_match_shape(g, vs, axis, keepdims)  # Avoid division by zero.
    if num_reps <= 1:
        return g_repeated * 0.0
    else:
        g_repeated, num_reps = repeat_to_match_shape(g / ans, vs, axis, keepdims)
        x_minus_mean = anp.conj(x - anp.mean(x, axis=axis, keepdims=True))
        return g_repeated * x_minus_mean / (num_reps - ddof)
anp.std.defvjp(grad_np_std)

def grad_chooser(g, ans, vs, gvs, x, axis=None, keepdims=None):
    """Builds gradient of functions that choose a single item, such as min or max."""
    g_repeated, _ = repeat_to_match_shape(g, vs, axis, keepdims)
    argmax_locations = x == repeat_to_match_shape(ans, vs, axis, keepdims)[0]
    return g_repeated * argmax_locations \
        / onp.sum(argmax_locations, axis=axis, keepdims=True)

anp.max.defvjp(grad_chooser)
anp.min.defvjp(grad_chooser)
anp.amax.defvjp(grad_chooser)
anp.amin.defvjp(grad_chooser)

def reverse_axis(x, axis):
    x = x.swapaxes(axis, 0)
    x = x[::-1,...]
    return x.swapaxes(0, axis)

def grad_np_cumsum(g, ans, vs, gvs, x, axis=None):
    if axis:
        return reverse_axis(anp.cumsum(reverse_axis(g, axis), axis), axis)
    else:
        return anp.reshape(anp.cumsum(g[::-1], axis)[::-1], x.shape)
anp.cumsum.defvjp(grad_np_cumsum)

def grad_inner(argnum, g, ans, vs, gvs, A, B):
    if anp.ndim(A) == 0 or anp.ndim(B) == 0:
        axes = ([], [])
    else:
        axes = ([A.ndim - 1], [B.ndim - 1])
    return grad_tensordot(argnum, g, ans, vs, gvs, A, B, axes=axes)
anp.inner.defvjps(grad_inner, [0, 1])

def grad_matmul(argnum, g, ans, vs, gvs, A, B):
    if anp.ndim(A) == 0 or anp.ndim(B) == 0:
        raise ValueError("Scalar operands are not allowed, use '*' instead")
    elif anp.ndim(A) == 1 or anp.ndim(B) == 1 or (anp.ndim(A) == 2 and anp.ndim(B) == 2):
        axes = ([A.ndim - 1], [max(0, B.ndim - 2)])
        return grad_tensordot(argnum, g, ans, vs, gvs, A, B, axes=axes)
    else:
        return grad_einsum(argnum + 1, g, ans, vs, gvs, ("...ij,...jk->...ik", A, B), None)
anp.matmul.defvjps(grad_matmul, [0, 1])

def grad_dot(argnum, g, ans, vs, gvs, A, B):
    A_ndim, B_ndim = anp.ndim(A), anp.ndim(B)
    if A_ndim == 0 or B_ndim == 0:
        axes = ([], [])
    else:
        axes = ([A_ndim - 1], [max(0, B_ndim - 2)])
    return grad_tensordot(argnum, g, ans, vs, gvs, A, B, axes=axes)
anp.dot.defvjps(grad_dot, [0, 1])

def grad_tensordot(argnum, g, ans, vs, gvs, A, B, axes=2):
    if anp.size(A) == anp.size(B) == 0:
        return g * B if argnum == 0 else g * A

    A_ndim = anp.ndim(A)
    g_axes = onp.arange(anp.ndim(g))
    if type(axes) is int:
        axes = max(axes, 0)
        if argnum == 0:
            B_axes = onp.arange(anp.ndim(B))
            return anp.tensordot(g, B, [g_axes[A_ndim-axes:], B_axes[axes:]])
        else:
            A_axes = onp.arange(A_ndim)
            return anp.tensordot(A, g, [A_axes[:A_ndim-axes], g_axes[:A_ndim-axes]])
    elif type(axes[0]) is int:
        B_ndim = anp.ndim(B)
        axes = [axes[0] % A_ndim, axes[1] % B_ndim]
        if argnum == 0:
            B_axes = onp.arange(B_ndim)
            return anp.tensordot(g, B, [g_axes[A_ndim-1:], onp.delete(B_axes, axes[1])])
        else:
            A_axes = onp.arange(A_ndim)
            return anp.tensordot(A, g, [onp.delete(A_axes, axes[0]), g_axes[:A_ndim-1]])
    else:
        B_ndim = anp.ndim(B)
        A_axes = onp.arange(A_ndim)
        B_axes = onp.arange(B_ndim)
        summed_axes = [onp.asarray(axes[0]) % A_ndim,
                       onp.asarray(axes[1]) % B_ndim]
        other_axes  = [onp.delete(A_axes, summed_axes[0]),
                       onp.delete(B_axes, summed_axes[1])]
        if argnum == 0:
            out = anp.tensordot(g, B, [g_axes[len(other_axes[0]):], other_axes[1]])
            perm = onp.argsort(onp.concatenate(
                (other_axes[0], summed_axes[0][onp.argsort(summed_axes[1])])))
            return anp.transpose(out, perm)
        else:
            out = anp.tensordot(A, g, [other_axes[0], g_axes[:len(other_axes[0])]])
            perm = onp.argsort(onp.concatenate(
                (summed_axes[1][onp.argsort(summed_axes[0])], other_axes[1])))
            return anp.transpose(out, perm)
anp.tensordot.defvjps(grad_tensordot, [0, 1])

anp.outer.defvjp(lambda g, ans, vs, gvs, a, b : anp.dot(g, b.T))
anp.outer.defvjp(lambda g, ans, vs, gvs, a, b : anp.dot(a.T, g), argnum=1)

def grad_concatenate_args(argnum, g, ans, vs, gvs, axis_args, kwargs):
    axis, args = axis_args[0], axis_args[1:]
    sizes = [a.shape[axis] for a in args[:argnum]]
    start = sum(sizes[:-1])
    idxs = [slice(None)] * ans.ndim
    idxs[axis] = slice(start, start + sizes[-1])
    return take(g, idxs)
anp.concatenate_args.vjp = grad_concatenate_args

def wrapped_reshape(x, *args, **kwargs):
    # The reshape method can be called like A.reshape((5,4)) or A.reshape(5,4).
    # The reshape function doesn't support both ways, so we have to wrap it.
    if isinstance(args[0], int):
        return anp.reshape(x, args, **kwargs)
    else:
        return anp.reshape(x, *args, **kwargs)
setattr(ArrayNode, 'reshape', wrapped_reshape)

def grad_sort(g, ans, vs, gvs, x, axis=-1, kind='quicksort', order=None):
    #TODO: Cast input with np.asanyarray()
    if len(x.shape) > 1:
        raise NotImplementedError(
            "Gradient of sort not implemented for multi-dimensional arrays.")
    sort_perm = anp.argsort(x, axis, kind, order)
    return unpermuter(g, sort_perm)
anp.sort.defvjp(grad_sort)
anp.msort.defvjp(grad_sort)  # Until multi-D is allowed, these are the same.

def grad_partition(g, ans, vs, gvs, x, kth, axis=-1, kind='introselect', order=None):
    #TODO: Cast input with np.asanyarray()
    if len(x.shape) > 1:
        raise NotImplementedError(
            "Gradient of partition not implemented for multi-dimensional arrays.")
    partition_perm = anp.argpartition(x, kth, axis, kind, order)
    return unpermuter(g, partition_perm)
anp.partition.defvjp(grad_partition)

def unpermuter(g, permutation):
    unsort = anp.zeros(len(permutation), dtype=int)
    unsort[permutation] = list(range(len(permutation)))
    return g[unsort]

def grad_reshape_list(g, ans, vs, gvs, *arys):
    if len(arys) > 1:
        raise NotImplementedError("Can't handle multiple arguments yet.")
    return anp.reshape(g, anp.shape(arys[0]))
anp.atleast_1d.defvjp(grad_reshape_list)
anp.atleast_2d.defvjp(grad_reshape_list)
anp.atleast_3d.defvjp(grad_reshape_list)

def grad_einsum(argnum, g, ans, vs, gvs, operands, kwargs):
    if isinstance(operands[0], string_types):  # using "ijk" convention.
        in_subs, out_subs, _ = _parse_einsum_input(tuple(map(getval, operands)))
        string, operands = operands[0], operands[1:]

        in_subs_list = in_subs.split(',')
        op_num = argnum - 1
        subs_wrt = in_subs_list[op_num]
        rest_of_ops = operands[:op_num] + operands[op_num+1:]
        rest_of_subs = in_subs_list[:op_num] + in_subs_list[op_num+1:]

        # subscripts that only appear in subs_wrt (and not in other subscript lists
        # or in the output) are implicitly being summed out, as if contracted
        # against a tensor of ones. we make that tensor of ones explicit to handle
        # the necessary vjp broadcasting inside einsum.
        other_named_subs = set(''.join([out_subs] + rest_of_subs))
        naked_summed = [(i, sub) for i, sub in enumerate(subs_wrt)
                        if sub not in other_named_subs]
        if naked_summed:
            naked_summed_dims, ones_subs = zip(*naked_summed)
            ones_subs = ''.join(ones_subs)
            ones = onp.ones(onp.array(operands[op_num].shape)[list(naked_summed_dims)])
            new_input_subs = ','.join([out_subs, ones_subs] + rest_of_subs)
            new_operands = (g, ones) + rest_of_ops
        else:
            new_input_subs = ','.join([out_subs] + rest_of_subs)
            new_operands = (g,) + rest_of_ops

        new_subscripts = new_input_subs + '->' + subs_wrt
        return unbroadcast(vs, gvs, anp.einsum(new_subscripts, *new_operands))
    else:  # using (op0, sublist0, op1, sublist1, ..., sublistout) convention
        if len(operands) % 2 == 0:
            raise NotImplementedError("Need sublistout argument")
        operands = list(operands)
        rest_of_ops = [operands[-1]] + operands[:argnum] + \
                operands[(argnum+2):-1] + [operands[argnum+1]]
        return unbroadcast_einsum(vs, gvs, anp.einsum(g, *rest_of_ops), operands[argnum + 1])
anp.einsum.vjp = grad_einsum

@primitive
def make_diagonal(D, offset=0, axis1=0, axis2=1):
    # Numpy doesn't offer a complement to np.diagonal: a function to create new
    # diagonal arrays with extra dimensions. We need such a function for the
    # gradient of np.diagonal and it's also quite handy to have. So here it is.
    if not (offset==0 and axis1==-1 and axis2==-2):
        raise NotImplementedError("Currently make_diagonal only supports offset=0, axis1=-1, axis2=-2")

    # We use a trick: calling np.diagonal returns a view on the original array,
    # so we can modify it in-place. (only valid for numpy version >= 1.10.)
    new_array = onp.zeros(D.shape + (D.shape[-1],))
    new_array_diag = onp.diagonal(new_array, offset=0, axis1=-1, axis2=-2)
    new_array_diag.flags.writeable = True
    new_array_diag[:] = D
    return new_array

anp.make_diagonal = make_diagonal
anp.diagonal.defvjp(
    lambda g, ans, vs, gvs, A, offset=0, axis1=0, axis2=1 :
    anp.make_diagonal(g, offset, axis1, axis2))
anp.make_diagonal.defvjp(
    lambda g, ans, vs, gvs, D, offset=0, axis1=0, axis2=1 :
    anp.diagonal(g, offset, axis1, axis2))

def match_complex(vs, x):
    x_iscomplex = vspace(x).iscomplex
    if x_iscomplex and not vs.iscomplex:
        return anp.real(x)
    elif not x_iscomplex and vs.iscomplex:
        return x + 0j
    else:
        return x

def unbroadcast(vs, gvs, result, broadcast_idx=0):
    while anp.ndim(result) > len(vs.shape):
        result = anp.sum(result, axis=broadcast_idx)
    for axis, size in enumerate(vs.shape):
        if size == 1:
            result = anp.sum(result, axis=axis, keepdims=True)
    if gvs.iscomplex and not vs.iscomplex:
        result = anp.real(result)
    return result

def unbroadcast_einsum(vs, gvs, result, subscript):
    if Ellipsis not in subscript:
        return result
    elif subscript[0] == Ellipsis:
        return unbroadcast(vs, gvs, result, 0)
    elif subscript[-1] == Ellipsis:
        return unbroadcast(vs, gvs, result, -1)
    else:
        return unbroadcast(vs, gvs, result, subscript.index(Ellipsis))

def balanced_eq(x, z, y):
    return (x == z) / (1.0 + (x == y))

def replace_zero(x, val):
    return anp.where(x, x, val)



from __future__ import absolute_import
import sys
import types
import numpy as np
import numpy.random as npr
from functools import partial
import warnings
from .errors import defgrad_deprecated

def make_vjp(fun, argnum=0):
    def vjp_maker(*args, **kwargs):
        start_node, end_node = forward_pass(fun, args, kwargs, argnum)
        if not isnode(end_node) or start_node not in end_node.progenitors:
            warnings.warn("Output seems independent of input.")
            def vjp(g): return start_node.vspace.zeros()
        else:
            def vjp(g): return backward_pass(g, end_node, start_node)
        return vjp, end_node
    return vjp_maker

def forward_pass(fun, args, kwargs, argnum=0):
    args = list(args)
    start_node = new_progenitor(args[argnum])
    args[argnum] = start_node
    active_progenitors.add(start_node)
    end_node = fun(*args, **kwargs)
    active_progenitors.remove(start_node)
    return start_node, end_node

def backward_pass(g, end_node, start_node):
    outgrads = {end_node : (g, False)}
    assert_vspace_match(outgrads[end_node][0], end_node.vspace, None)
    for node in toposort(end_node, start_node):
        cur_outgrad = outgrads.pop(node)
        function, args, kwargs, parents = node.recipe
        for argnum, parent in parents:
            outgrad = function.vjp(argnum, cur_outgrad[0], node,
                                   parent.vspace, node.vspace, args, kwargs)
            assert_vspace_match(outgrad, parent.vspace, function)
            outgrads[parent] = add_outgrads(parent.vspace, outgrads.get(parent), outgrad)
    return cur_outgrad[0]

def add_outgrads(vspace, prev_g_flagged, g):
    if prev_g_flagged is None:
        if type(getval(g)) == SparseObject:
            return primitive_mut_add(vspace, None, g), True
        else:
            return g, False
    else:
        prev_g, mutable = prev_g_flagged
        if mutable:
            return primitive_mut_add(vspace, prev_g, g), True
        else:
            prev_g_mutable = primitive_mut_add(vspace, None, prev_g)
            return primitive_mut_add(vspace, prev_g_mutable, g), True

active_progenitors = set()

class primitive(object):
    """
    Wraps a function so that its gradient can be specified and its invocation
    can be recorded. For examples, see the docs."""
    def __init__(self, fun):
        self.fun = fun
        self.vjps = {}
        self.zero_vjps = set()
        self.__name__ = fun.__name__
        self.__doc__ = fun.__doc__

    def __call__(self, *args, **kwargs):
        argvals = list(args)
        progenitors = set()
        parents = []
        for argnum, arg in enumerate(args):
            if isnode(arg):
                argvals[argnum] = arg.value
                if argnum in self.zero_vjps: continue
                parents.append((argnum, arg))
                progenitors.update(arg.progenitors & active_progenitors)

        result_value = self.fun(*argvals, **kwargs)
        if progenitors:
            return new_node(result_value, (self, args, kwargs, parents), progenitors)
        else:
            return result_value

    def vjp(self, argnum, outgrad, ans, vs, gvs, args, kwargs):
        try:
            return self.vjps[argnum](outgrad, ans, vs, gvs, *args, **kwargs)
        except KeyError:
            if self.vjps == {}:
                errstr = "Gradient of {0} not yet implemented."
            else:
                errstr = "Gradient of {0} w.r.t. arg number {1} not yet implemented."
            raise NotImplementedError(errstr.format(self.fun.__name__, argnum))

    def defvjp(self, vjpmaker, argnum=0):
        vjpmaker.__name__ = "VJP_{}_of_{}".format(argnum, self.__name__)
        self.vjps[argnum] = vjpmaker

    def defvjps(self, vjpmaker, argnums):
        for argnum in argnums:
            self.defvjp(partial(vjpmaker, argnum), argnum)

    def defvjp_is_zero(self, argnums=(0,)):
        for argnum in argnums:
            self.zero_vjps.add(argnum)

    def __repr__(self):
        return self.__name__

    if sys.version_info >= (3,):
        def __get__(self, obj, objtype):
            return types.MethodType(self, obj)
    else:
        def __get__(self, obj, objtype):
            return types.MethodType(self, obj, objtype)

    def defgrad(self, gradfun, argnum=0):
        warnings.warn(defgrad_deprecated)
        def vjp(g, ans, vs, gvs, *args, **kwargs):
            return gradfun(ans, *args, **kwargs)(g)
        self.defvjp(vjp, argnum)

class nograd_primitive(primitive):
    def __call__(self, *args, **kwargs):
        argvals = map(getval, args)
        return self.fun(*argvals, **kwargs)

@primitive
def primitive_mut_add(vspace, x_prev, x_new):
    if x_prev is None:
        x_prev = vspace.zeros()
    if type(x_new) == SparseObject:
        return x_new.mut_add(x_prev)
    else:
        return vspace.mut_add(x_prev, x_new)
primitive_mut_add.vjp = lambda argnum, g, *args : g

def new_progenitor(x):
    if isnode(x):
        node = new_node(x.value, (identity, (x,), {}, [(0, x)]), x.progenitors)
    else:
        node = new_node(x,       (identity, (x,), {}, []      ), set())
    node.progenitors = node.progenitors | {node}
    return node

@primitive
def identity(x) : return x
identity.defvjp(lambda g, ans, vs, gvs, x : g)

class Node(object):
    __slots__ = ['value', 'recipe', 'progenitors', 'vspace']

    def __init__(self, value, recipe, progenitors):
        self.value = value
        self.recipe = recipe
        self.progenitors = progenitors
        self.vspace = vspace(value)

    def __bool__(self):
        return bool(self.value)

    __nonzero__ = __bool__

    def __str__(self):
        return "Autograd {0} with value {1} and {2} progenitors(s)".format(
            type(self).__name__, str(self.value), len(self.progenitors))

def toposort(end_node, start_node):
    def relevant_parents(node):
        return [parent for _, parent in node.recipe[3] if start_node in parent.progenitors]

    child_counts = {}
    stack = [end_node]
    while stack:
        node = stack.pop()
        if node in child_counts:
            child_counts[node] += 1
        else:
            child_counts[node] = 1
            stack.extend(relevant_parents(node))

    childless_nodes = [end_node]
    while childless_nodes:
        node = childless_nodes.pop()
        yield node
        for parent in relevant_parents(node):
            if child_counts[parent] == 1:
                childless_nodes.append(parent)
            else:
                child_counts[parent] -= 1

class VSpace(object):
    __slots__ = []
    iscomplex = False
    def __init__(self, value):
        pass

    def zeros(self):
        assert False

    def ones(self):
        assert False

    def standard_basis(self):
        assert False

    def mut_add(self, x, y):
        x += y
        return x

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __repr__(self):
        return "{}_{}".format(type(self).__name__, self.__dict__)

    def examples(self):
        # Used for testing only
        N = self.size
        unit_vect = np.zeros(N)
        unit_vect[npr.randint(N)] = 1.0
        unit_vect = self.unflatten(unit_vect)
        rand_vect = npr.randn(N)
        return [self.zeros(), self.unflatten(npr.randn(N))]

def vspace_flatten(value, covector=False):
    return vspace(value).flatten(value, covector)

node_type_mappings = {}
vspace_mappings = {}
node_types = set()
def register_node(node_type, value_type):
    node_types.add(node_type)
    node_type_mappings[value_type] = node_type
    node_type_mappings[node_type] = node_type

def register_vspace(vspace_maker, value_type):
    vspace_mappings[value_type] = vspace_maker

def new_node(value, recipe, progenitors):
    try:
        return node_type_mappings[type(value)](value, recipe, progenitors)
    except KeyError:
        raise TypeError("Can't differentiate w.r.t. type {}".format(type(value)))

def vspace(value):
    try:
        return vspace_mappings[type(value)](value)
    except KeyError:
        if isnode(value):
            return value.vspace
        else:
            raise TypeError("Can't find vspace for type {}".format(type(value)))

class SparseObject(object):
    __slots__ = ['vs', 'mut_add']
    def __init__(self, vs, mut_add):
        self.vs = vs
        self.mut_add = mut_add

register_vspace(lambda x : x.vs, SparseObject)
register_node(Node, SparseObject)

def assert_vspace_match(x, expected_vspace, fun):
    assert expected_vspace == vspace(x), \
        "\nGrad of {} returned unexpected vector space" \
        "\nVector space is {}" \
        "\nExpected        {}".format(fun, vspace(x), expected_vspace)

isnode = lambda x: type(x) in node_types
getval = lambda x: x.value if isnode(x) else x

def unbox_if_possible(node):
    if isnode(node) and not active_progenitors.intersection(node.progenitors):
        return node.value
    return node



from __future__ import division
import scipy.linalg

import autograd.numpy as anp
from autograd.numpy.numpy_wrapper import wrap_namespace

wrap_namespace(scipy.linalg.__dict__, globals())  # populates module namespace

sqrtm.defvjp(lambda g, ans, vs, gvs, A, **kwargs: solve_lyapunov(ans, g))

def _flip(a, trans):
    if anp.iscomplexobj(a):
        return 'H' if trans in ('N', 0) else 'N'
    else:
        return 'T' if trans in ('N', 0) else 'N'

def grad_solve_triangular(g, ans, vs, gvs, a, b, trans=0, lower=False, **kwargs):
    tri = anp.tril if (lower ^ (_flip(a, trans) == 'N')) else anp.triu
    transpose = lambda x: x if _flip(a, trans) != 'N' else x.T
    al2d = lambda x: x if x.ndim > 1 else x[...,None]
    v = al2d(solve_triangular(a, g, trans=_flip(a, trans), lower=lower))
    return -transpose(tri(anp.dot(v, al2d(ans).T)))

solve_triangular.defvjp(grad_solve_triangular)
solve_triangular.defvjp(lambda g, ans, vs, gvs, a, b, trans=0, lower=False, **kwargs:
    solve_triangular(a, g, trans=_flip(a, trans), lower=lower), argnum=1)



from __future__ import absolute_import
from autograd.core import (primitive, Node, VSpace, register_node, vspace,
                           register_vspace, SparseObject)
from builtins import zip
from future.utils import iteritems
from functools import partial
import autograd.numpy as np

class SequenceNode(Node):
    __slots__ = []
    def __getitem__(self, idx): return sequence_take(self, idx)
    def __len__(self): return len(self.value)
    def __add__(self, other): return sequence_extend_right(self, *other)
    def __radd__(self, other): return sequence_extend_left(self, *other)

register_node(SequenceNode, tuple)
register_node(SequenceNode, list)

@primitive
def sequence_take(A, idx):
    return A[idx]
def grad_sequence_take(g, ans, vs, gvs, A, idx):
    return sequence_untake(g, idx, vs)
sequence_take.defvjp(grad_sequence_take)

@primitive
def sequence_extend_right(seq, *elts):
    return seq + type(seq)(elts)
def grad_sequence_extend_right(argnum, g, ans, vs, gvs, args, kwargs):
    seq, elts = args[0], args[1:]
    return g[:len(seq)] if argnum == 0 else g[len(seq) + argnum - 1]
sequence_extend_right.vjp = grad_sequence_extend_right

@primitive
def sequence_extend_left(seq, *elts):
    return type(seq)(elts) + seq
def grad_sequence_extend_left(argnum, g, ans, vs, gvs, args, kwargs):
    seq, elts = args[0], args[1:]
    return g[len(elts):] if argnum == 0 else g[argnum - 1]
sequence_extend_left.vjp = grad_sequence_extend_left

@primitive
def sequence_untake(x, idx, vs):
    if isinstance(idx, int):
        accum = lambda result: vs.shape[idx].mut_add(result, x)
    else:
        accum = lambda result: [elt_vs.mut_add(a, b)
                                for elt_vs, a, b in zip(vs.shape[idx], result, x)]
    def mut_add(A):
        result = list(A)
        result[idx] = accum(result[idx])
        return vs.sequence_type(result)
    return SparseObject(vs, mut_add)
sequence_untake.defvjp(lambda g, ans, vs, gvs, x, idx, _: sequence_take(g, idx))
sequence_untake.defvjp_is_zero(argnums=(1, 2))

@primitive
def make_sequence(sequence_type, *args):
    return sequence_type(args)
make_sequence.vjp = lambda argnum, g, sequence_type, *args: g[argnum - 1]
make_tuple = partial(make_sequence, tuple)
make_list  = partial(make_sequence, list)

class SequenceVSpace(VSpace):
    def __init__(self, value):
        self.shape = [vspace(x) for x in value]
        self.size = sum(s.size for s in self.shape)
        self.sequence_type = type(value)
        assert self.sequence_type in (tuple, list)

    def zeros(self):
        return self.sequence_type(x.zeros() for x in self.shape)

    def mut_add(self, xs, ys):
        return self.sequence_type(vs.mut_add(x, y)
                                  for vs, x, y in zip(self.shape, xs, ys))

    def flatten(self, value, covector=False):
        if self.shape:
            return np.concatenate(
                [vs.flatten(v, covector) for vs, v in zip(self.shape, value)])
        else:
            return np.zeros((0,))

    def unflatten(self, value, covector=False):
        result = []
        start = 0
        for vs in self.shape:
            N = vs.size
            result.append(vs.unflatten(value[start:start + N], covector))
            start += N
        return self.sequence_type(result)

register_vspace(SequenceVSpace, list)
register_vspace(SequenceVSpace, tuple)

class DictNode(Node):
    __slots__ = []
    def __getitem__(self, idx): return dict_take(self, idx)
    def __len__(self): return len(self.value)
    def __iter__(self): return self.value.__iter__()
    def items(self): return list(self.iteritems())
    def keys(self): return list(self.iterkeys())
    def values(self): return list(self.itervalues())
    def iteritems(self): return ((k, self[k]) for k in self)
    def iterkeys(self): return iter(self)
    def itervalues(self): return (self[k] for k in self)

register_node(DictNode, dict)

@primitive
def dict_take(A, idx):
    return A[idx]
def grad_dict_take(g, ans, vs, gvs, A, idx):
    return dict_untake(g, idx, vs)
dict_take.defvjp(grad_dict_take)

@primitive
def dict_untake(x, idx, vs):
    def mut_add(A):
         A[idx] = vs.shape[idx].mut_add(A[idx], x)
         return A
    return SparseObject(vs, mut_add)
dict_untake.defvjp(lambda g, ans, vs, gvs, x, idx, _: dict_take(g, idx))
dict_untake.defvjp_is_zero(argnums=(1, 2))

def make_dict(pairs):
    keys, vals = zip(*pairs)
    return _make_dict(make_list(*keys), make_list(*vals))
@primitive
def _make_dict(keys, vals):
    return dict(zip(keys, vals))
_make_dict.defvjp(lambda g, ans, vs, gvs, keys, vals: make_list(*[g[key] for key in keys]),
                  argnum=1)

class DictVSpace(VSpace):
    def __init__(self, value):
        self.shape = {k : vspace(v) for k, v in iteritems(value)}
        self.size  = sum(s.size for s in self.shape.values())
    def zeros(self):
        return {k : v.zeros() for k, v in iteritems(self.shape)}
    def mut_add(self, xs, ys):
        return {k : v.mut_add(xs[k], ys[k])
                for k, v in iteritems(self.shape)}
    def flatten(self, value, covector=False):
        if self.shape:
            return np.concatenate(
                [s.flatten(value[k], covector)
                 for k, s in sorted(iteritems(self.shape))])
        else:
            return np.zeros((0,))

    def unflatten(self, value, covector=False):
        result = {}
        start = 0
        for k, s in sorted(iteritems(self.shape)):
            N = s.size
            result[k] = s.unflatten(value[start:start + N], covector)
            start += N
        return result

register_vspace(DictVSpace, dict)



from __future__ import absolute_import
from . import numpy_wrapper
from . import numpy_grads
from . import numpy_extra
from .numpy_wrapper import *
from . import linalg
from . import fft
from . import random



"""Some standard gradient-based stochastic optimizers.

These are just standard routines that don't make any use of autograd,
though you could take gradients of these functions too if you want
to do meta-optimization.

These routines can optimize functions whose inputs are structured
objects, such as dicts of numpy arrays."""
from __future__ import absolute_import

import autograd.numpy as np
from autograd.util import flatten_func
from builtins import range


def sgd(grad, init_params, callback=None, num_iters=200, step_size=0.1, mass=0.9):
    """Stochastic gradient descent with momentum.
    grad() must have signature grad(x, i), where i is the iteration number."""
    flattened_grad, unflatten, x = flatten_func(grad, init_params)

    velocity = np.zeros(len(x))
    for i in range(num_iters):
        g = flattened_grad(x, i)
        if callback: callback(unflatten(x), i, unflatten(g))
        velocity = mass * velocity - (1.0 - mass) * g
        x = x + step_size * velocity
    return unflatten(x)

def rmsprop(grad, init_params, callback=None, num_iters=100,
            step_size=0.1, gamma=0.9, eps=10**-8):
    """Root mean squared prop: See Adagrad paper for details."""
    flattened_grad, unflatten, x = flatten_func(grad, init_params)

    avg_sq_grad = np.ones(len(x))
    for i in range(num_iters):
        g = flattened_grad(x, i)
        if callback: callback(unflatten(x), i, unflatten(g))
        avg_sq_grad = avg_sq_grad * gamma + g**2 * (1 - gamma)
        x = x - step_size * g/(np.sqrt(avg_sq_grad) + eps)
    return unflatten(x)

def adam(grad, init_params, callback=None, num_iters=100,
         step_size=0.001, b1=0.9, b2=0.999, eps=10**-8):
    """Adam as described in http://arxiv.org/pdf/1412.6980.pdf.
    It's basically RMSprop with momentum and some correction terms."""
    flattened_grad, unflatten, x = flatten_func(grad, init_params)

    m = np.zeros(len(x))
    v = np.zeros(len(x))
    for i in range(num_iters):
        g = flattened_grad(x, i)
        if callback: callback(unflatten(x), i, unflatten(g))
        m = (1 - b1) * g      + b1 * m  # First  moment estimate.
        v = (1 - b2) * (g**2) + b2 * v  # Second moment estimate.
        mhat = m / (1 - b1**(i + 1))    # Bias correction.
        vhat = v / (1 - b2**(i + 1))
        x = x - step_size*mhat/(np.sqrt(vhat) + eps)
    return unflatten(x)



from __future__ import absolute_import
import scipy.misc

from autograd.core import primitive
import autograd.numpy as anp
from autograd.numpy.numpy_grads import repeat_to_match_shape

logsumexp = primitive(scipy.misc.logsumexp)

def make_grad_logsumexp(g, ans, vs, gvs, x, axis=None, b=1.0, keepdims=False):
    g_repeated,   _ = repeat_to_match_shape(g,   vs, axis, keepdims)
    ans_repeated, _ = repeat_to_match_shape(ans, vs, axis, keepdims)
    return g_repeated * b * anp.exp(x - ans_repeated)

logsumexp.defvjp(make_grad_logsumexp)



from __future__ import absolute_import
import numpy as np

from autograd.core import (Node, VSpace, SparseObject, primitive,
                           register_node, register_vspace)
from . import numpy_wrapper as anp

@primitive
def take(A, idx):
    return A[idx]
def grad_take(g, ans, vs, gvs, A, idx):
    return untake(g, idx, vs)
take.defvjp(grad_take)

@primitive
def untake(x, idx, vs):
    def mut_add(A):
        np.add.at(A, idx, x)
        return A
    return SparseObject(vs, mut_add)
untake.defvjp(lambda g, ans, vs, gvs, x, idx, _: take(g, idx))
untake.defvjp_is_zero(argnums=(1, 2))

Node.__array_priority__ = 90.0

class ArrayNode(Node):
    __slots__ = []
    __getitem__ = take
    __array_priority__ = 100.0

    # Constants w.r.t float data just pass though
    shape = property(lambda self: self.value.shape)
    ndim  = property(lambda self: self.value.ndim)
    size  = property(lambda self: self.value.size)
    dtype = property(lambda self: self.value.dtype)
    T = property(lambda self: anp.transpose(self))

    def __len__(self):
        return len(self.value)

    def __neg__(self): return anp.negative(self)
    def __add__(self, other): return anp.add(     self, other)
    def __sub__(self, other): return anp.subtract(self, other)
    def __mul__(self, other): return anp.multiply(self, other)
    def __pow__(self, other): return anp.power   (self, other)
    def __div__(self, other): return anp.divide(  self, other)
    def __mod__(self, other): return anp.mod(     self, other)
    def __truediv__(self, other): return anp.true_divide(self, other)
    def __matmul__(self, other): return anp.matmul(self, other)
    def __radd__(self, other): return anp.add(     other, self)
    def __rsub__(self, other): return anp.subtract(other, self)
    def __rmul__(self, other): return anp.multiply(other, self)
    def __rpow__(self, other): return anp.power(   other, self)
    def __rdiv__(self, other): return anp.divide(  other, self)
    def __rmod__(self, other): return anp.mod(     other, self)
    def __rtruediv__(self, other): return anp.true_divide(other, self)
    def __rmatmul__(self, other): return anp.matmul(other, self)
    def __eq__(self, other): return anp.equal(self, other)
    def __ne__(self, other): return anp.not_equal(self, other)
    def __gt__(self, other): return anp.greater(self, other)
    def __ge__(self, other): return anp.greater_equal(self, other)
    def __lt__(self, other): return anp.less(self, other)
    def __le__(self, other): return anp.less_equal(self, other)
    def __abs__(self): return anp.abs(self)
    def __hash__(self): return id(self)

class ArrayVSpace(VSpace):
    def __init__(self, value):
        value = np.array(value, copy=False)
        self.shape = value.shape
        self.size  = value.size
        self.dtype = value.dtype
        self.ndim  = value.ndim
        self.scalartype = float

    def zeros(self):
        return np.zeros(self.shape, dtype=self.dtype)

    def ones(self):
        return np.ones(self.shape, dtype=self.dtype)

    def standard_basis(self):
      for idxs in np.ndindex(*self.shape):
            vect = np.zeros(self.shape)
            vect[idxs] = 1
            yield vect

    def flatten(self, value, covector=False):
        return anp.ravel(value)

    def unflatten(self, value, covector=False):
        return value.reshape(self.shape)

    def examples(self):
        # many possible instantiations
        original_examples = super(ArrayVSpace, self).examples()
        if self.shape == ():
            np_scalar_examples = [ex[()] for ex in original_examples]
            py_scalar_examples = list(map(self.scalartype, np_scalar_examples))
            return original_examples + np_scalar_examples + py_scalar_examples
        else:
            return original_examples

class ComplexArrayVSpace(ArrayVSpace):
    iscomplex = True
    def __init__(self, value):
        super(ComplexArrayVSpace, self).__init__(value)
        self.size  = 2 * self.size
        self.scalartype = complex

    def flatten(self, value, covector=False):
        if covector:
            return anp.ravel(anp.stack([anp.real(value), - anp.imag(value)]))
        else:
            return anp.ravel(anp.stack([anp.real(value), anp.imag(value)]))

    def unflatten(self, value, covector=False):
        reshaped = anp.reshape(value, (2,) + self.shape)
        if covector:
            return anp.array(reshaped[0] - 1j * reshaped[1])
        else:
            return anp.array(reshaped[0] + 1j * reshaped[1])

register_node(ArrayNode, np.ndarray)
register_vspace(lambda x: ComplexArrayVSpace(x)
                if np.iscomplexobj(x)
                else ArrayVSpace(x), np.ndarray)
array_types = set([anp.ndarray, ArrayNode])

for type_ in [float, anp.float64, anp.float32, anp.float16]:
    register_node(ArrayNode, type_)
    register_vspace(ArrayVSpace, type_)

for type_ in [complex, anp.complex64, anp.complex128]:
    register_node(ArrayNode, type_)
    register_vspace(ComplexArrayVSpace, type_)

# These numpy.ndarray methods are just refs to an equivalent numpy function
nondiff_methods = ['all', 'any', 'argmax', 'argmin', 'argpartition',
                   'argsort', 'nonzero', 'searchsorted', 'round']
diff_methods = ['clip', 'compress', 'cumprod', 'cumsum', 'diagonal',
                'max', 'mean', 'min', 'prod', 'ptp', 'ravel', 'repeat',
                'reshape', 'squeeze', 'std', 'sum', 'swapaxes', 'take',
                'trace', 'transpose', 'var']
for method_name in nondiff_methods + diff_methods:
    setattr(ArrayNode, method_name, anp.__dict__[method_name])

# Flatten has no function, only a method.
setattr(ArrayNode, 'flatten', anp.__dict__['ravel'])



import sys
import re
from functools import wraps
from future.utils import raise_from, raise_

class AutogradHint(Exception):
    def __init__(self, message, subexception_type=None, subexception_val=None):
        self.message = message
        self.subexception_type = subexception_type
        self.subexception_val = subexception_val

    def __str__(self):
        if self.subexception_type:
            return '{message}\nSub-exception:\n{name}: {str}'.format(
                message=self.message,
                name=self.subexception_type.__name__,
                str=self.subexception_type(self.subexception_val))
        else:
            return self.message

common_errors = [
    ((TypeError, r'float() argument must be a string or a number'),
        "This error *might* be caused by assigning into arrays, which autograd doesn't support."),
    ((ValueError, r'setting an array element with a sequence'),
        "This error *might* be caused by assigning into arrays, which autograd doesn't support."),
    ((TypeError, r"got an unexpected keyword argument '(?:dtype)|(?:out)'" ),
        "This error *might* be caused by importing numpy instead of autograd.numpy. \n"
        "Check that you have 'import autograd.numpy as np' instead of 'import numpy as np'."),
    ((AttributeError, r"object has no attribute" ),
        "This error *might* be caused by importing numpy instead of autograd.numpy,"
        "or otherwise using a raw numpy function instead of the autograd-wrapped version. \n"
        "Check that you have 'import autograd.numpy as np' instead of 'import numpy as np'."),
]

defgrad_deprecated = \
'''
------------------------------
    defgrad is deprecated!
------------------------------
Use defvjp instead ("define vector-Jacobian product").
The interface is a little different - look at
autograd/numpy/numpy_grads.py for examples.
'''

def add_error_hints(fun):
  @wraps(fun)
  def wrapped(*args, **kwargs):
    try: return fun(*args, **kwargs)
    except Exception as e: add_extra_error_message(e)
  return wrapped

def check_common_errors(error_type, error_message):
    keys, vals = zip(*common_errors)
    matches = [error_type == key[0]
               and len(re.findall(key[1], error_message)) != 0
               for key in keys]
    num_matches = sum(matches)

    if num_matches == 1:
        return vals[matches.index(True)]

def add_extra_error_message(e):
    etype, value, traceback = sys.exc_info()
    extra_message = check_common_errors(type(e), str(e))

    if extra_message:
        if sys.version_info >= (3,):
            raise_from(AutogradHint(extra_message), e)
        else:
            raise_(AutogradHint, (extra_message, etype, value), traceback)
    raise_(etype, value, traceback)



from __future__ import absolute_import
from .core import primitive, make_vjp, getval
from . import container_types
from .container_types import make_tuple, make_list, make_dict
from .convenience_wrappers import (grad, multigrad, multigrad_dict, elementwise_grad,
                                   value_and_grad, grad_and_aux, hessian_tensor_product,
                                   hessian_vector_product, hessian, jacobian,
                                   tensor_jacobian_product, vector_jacobian_product,
                                   grad_named, checkpoint, make_hvp, value_and_multigrad,
                                   make_jvp, make_ggnvp)



from __future__ import absolute_import
from __future__ import print_function
from copy import copy
from builtins import range

import autograd.numpy as np
from autograd.convenience_wrappers import grad
from autograd.core import vspace, vspace_flatten, getval

EPS, RTOL, ATOL = 1e-4, 1e-4, 1e-6

def nd(f, *args):
    unary_f = lambda x : f(*x)
    return unary_nd(unary_f, args)

def unary_nd(f, x, eps=EPS):
    vs = vspace(x)
    nd_grad = np.zeros(vs.size)
    x_flat = vs.flatten(x)
    for d in range(vs.size):
        dx = np.zeros(vs.size)
        dx[d] = eps/2
        nd_grad[d] = (   f(vs.unflatten(x_flat + dx))
                       - f(vs.unflatten(x_flat - dx))  ) / eps
    return vs.unflatten(nd_grad, True)

def indexed_function(fun, arg, index):
    def partial_function(x):
        local_arg = copy(arg)
        if isinstance(local_arg, tuple):
            local_arg = local_arg[:index] + (x,) + local_arg[index+1:]
        elif isinstance(local_arg, list):
            local_arg = local_arg[:index] + [x] + local_arg[index+1:]
        else:
            local_arg[index] = x
        return fun(local_arg)
    return partial_function

def check_equivalent(A, B, rtol=RTOL, atol=ATOL):
    A_vspace = vspace(A)
    B_vspace = vspace(B)
    A_flat = vspace_flatten(A)
    B_flat = vspace_flatten(B)
    assert A_vspace == B_vspace, \
      "VSpace mismatch:\nanalytic: {}\nnumeric: {}".format(A_vspace, B_vspace)
    assert np.allclose(vspace_flatten(A), vspace_flatten(B), rtol=rtol, atol=atol), \
        "Diffs are:\n{}.\nanalytic is:\n{}.\nnumeric is:\n{}.".format(
            A_flat - B_flat, A_flat, B_flat)

def check_grads(fun, *args):
    if not args:
        raise Exception("No args given")
    exact = tuple([grad(fun, i)(*args) for i in range(len(args))])
    args = [float(x) if isinstance(x, int) else x for x in args]
    numeric = nd(fun, *args)
    check_equivalent(exact, numeric)

def to_scalar(x):
    if isinstance(getval(x), list)  or isinstance(getval(x), tuple):
        return sum([to_scalar(item) for item in x])
    return np.sum(np.real(np.sin(x)))

def quick_grad_check(fun, arg0, extra_args=(), kwargs={}, verbose=True,
                     eps=EPS, rtol=RTOL, atol=ATOL, rs=None):
    """Checks the gradient of a function (w.r.t. to its first arg) in a random direction"""

    if verbose:
        print("Checking gradient of {0} at {1}".format(fun, arg0))

    if rs is None:
        rs = np.random.RandomState()

    random_dir = rs.standard_normal(np.shape(arg0))
    random_dir = random_dir / np.sqrt(np.sum(random_dir * random_dir))
    unary_fun = lambda x : fun(arg0 + x * random_dir, *extra_args, **kwargs)
    numeric_grad = unary_nd(unary_fun, 0.0, eps=eps)

    analytic_grad = np.sum(grad(fun)(arg0, *extra_args, **kwargs) * random_dir)

    assert np.allclose(numeric_grad, analytic_grad, rtol=rtol, atol=atol), \
        "Check failed! nd={0}, ad={1}".format(numeric_grad, analytic_grad)

    if verbose:
        print("Gradient projection OK (numeric grad: {0}, analytic grad: {1})".
              format(numeric_grad, analytic_grad))

def flatten(value):
    """Flattens any nesting of tuples, arrays, or dicts.
       Returns 1D numpy array and an unflatten function.
       Doesn't preserve mixed numeric types (e.g. floats and ints).
       Assumes dict keys are sortable."""
    vs = vspace(value)
    return vs.flatten(value), vs.unflatten

def flatten_func(func, example):
    """Flattens both the inputs to a function, and the outputs."""
    flattened_example, unflatten = flatten(example)

    def flattened_func(flattened_params, *args, **kwargs):
        output = func(unflatten(flattened_params), *args, **kwargs)
        flattened_output, _ = flatten(output)
        return flattened_output
    return flattened_func, unflatten, flattened_example



"""Gradients of the univariate t distribution."""
from __future__ import absolute_import
import scipy.stats
import autograd.numpy as np

from autograd.core import primitive
from autograd.numpy.numpy_grads import unbroadcast
from autograd.scipy.special import psi

pdf = primitive(scipy.stats.t.pdf)
cdf = primitive(scipy.stats.t.cdf)
logpdf = primitive(scipy.stats.t.logpdf)
logcdf = primitive(scipy.stats.t.logcdf)

def grad_tlogpdf_diff(diff, df):
    return -diff * (1.0 + df) / (diff**2 + df)
def grad_tlogpdf_x(x, df, loc, scale):
    return grad_tlogpdf_diff((x - loc) / scale, df) / scale
def grad_tlogpdf_loc(x, df, loc, scale):
    return -grad_tlogpdf_diff((x - loc) / scale, df) / scale
def grad_tlogpdf_scale(x, df, loc, scale):
    diff = x - loc
    return -(df * (scale**2 - diff**2))/(scale * (df * scale**2 + diff**2))
def grad_tlogpdf_df(x, df, loc, scale):
    y = (x - loc)/scale
    return 0.5 * ((y**2 * (df+1))/(df * (y**2 + df)) - np.log(y**2 / df + 1) - 1.0/df -psi(df/2.0) + psi((df + 1)/2.0))

pdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * ans * grad_tlogpdf_x(    x, df, loc, scale)), argnum=0)
pdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * ans * grad_tlogpdf_df(   x, df, loc, scale)), argnum=1)
pdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * ans * grad_tlogpdf_loc(  x, df, loc, scale)), argnum=2)
pdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * ans * grad_tlogpdf_scale(x, df, loc, scale)), argnum=3)

cdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs,  g * pdf(x, df, loc, scale)), argnum=0)
cdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, -g * pdf(x, df, loc, scale)), argnum=2)
# What is the gradient of the cdf wrt the degrees of freedom or scale?  No one knows.

logpdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * grad_tlogpdf_x(    x, df, loc, scale)), argnum=0)
logpdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * grad_tlogpdf_df(   x, df, loc, scale)), argnum=1)
logpdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * grad_tlogpdf_loc(  x, df, loc, scale)), argnum=2)
logpdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, g * grad_tlogpdf_scale(x, df, loc, scale)), argnum=3)

logcdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs,  g * np.exp(logpdf(x, df, loc, scale) - logcdf(x, df, loc, scale))), argnum=0)
logcdf.defvjp(lambda g, ans, vs, gvs, x, df, loc=0.0, scale=1.0: unbroadcast(vs, gvs, -g * np.exp(logpdf(x, df, loc, scale) - logcdf(x, df, loc, scale))), argnum=2)



from __future__ import absolute_import
from functools import partial
import numpy.linalg as npla
from .numpy_wrapper import wrap_namespace
from . import numpy_wrapper as anp

wrap_namespace(npla.__dict__, globals())

# Some formulas are from
# "An extended collection of matrix derivative results
#  for forward and reverse mode algorithmic differentiation"
# by Mike Giles
# https://people.maths.ox.ac.uk/gilesm/files/NA-08-01.pdf

# transpose by swapping last two dimensions
def T(x): return anp.swapaxes(x, -1, -2)

# add two dimensions to the end of x
def add2d(x): return anp.reshape(x, anp.shape(x) + (1, 1))

det.defvjp(lambda g, ans, vs, gvs, x: add2d(g) * add2d(ans) * T(inv(x)))
slogdet.defvjp(lambda g, ans, vs, gvs, x: add2d(g[1]) * T(inv(x)))

def grad_inv(g, ans, vs, gvs, x):
    dot = anp.dot if ans.ndim == 2 else partial(anp.einsum, '...ij,...jk->...ik')
    return -dot(dot(T(ans), g), T(ans))
inv.defvjp(grad_inv)

def grad_solve(argnum, g, ans, vs, gvs, a, b):
    updim = lambda x: x if x.ndim == a.ndim else x[...,None]
    dot = anp.dot if a.ndim == 2 else partial(anp.einsum, '...ij,...jk->...ik')
    if argnum == 0:
        return -dot(updim(solve(T(a), g)), T(updim(ans)))
    else:
        return solve(T(a), g)
solve.defvjps(grad_solve, [0, 1])

def grad_norm(g, ans, vs, gvs, x, ord=None, axis=None):
    def check_implemented():
        matrix_norm = (x.ndim == 2 and axis is None) or isinstance(axis, tuple)

        if matrix_norm:
            if not (ord is None or ord == 'fro' or ord == 'nuc'):
                raise NotImplementedError('Gradient of matrix norm not '
                                          'implemented for ord={}'.format(ord))
        elif not (ord is None or ord > 1):
            raise NotImplementedError('Gradient of norm not '
                                      'implemented for ord={}'.format(ord))

    if axis is None:
        expand = lambda a: a
    elif isinstance(axis, tuple):
        row_axis, col_axis = axis
        if row_axis > col_axis:
            row_axis = row_axis - 1
        expand = lambda a: anp.expand_dims(anp.expand_dims(a,
                                                   row_axis), col_axis)
    else:
        expand = lambda a: anp.expand_dims(a, axis=axis)

    if ord == 'nuc':
        if axis is None:
            roll = lambda a: a
            unroll = lambda a: a
        else:
            row_axis, col_axis = axis
            if row_axis > col_axis:
                row_axis = row_axis - 1
            # Roll matrix axes to the back
            roll = lambda a: anp.rollaxis(anp.rollaxis(a, col_axis, a.ndim),
                                          row_axis, a.ndim-1)
            # Roll matrix axes to their original position
            unroll = lambda a: anp.rollaxis(anp.rollaxis(a, a.ndim-2, row_axis),
                                            a.ndim-1, col_axis)

    check_implemented()
    if ord is None or ord == 2 or ord is 'fro':
        return expand(g / ans) * x
    elif ord == 'nuc':
        dot = anp.dot if x.ndim == 2 else partial(anp.einsum, '...ij,...jk->...ik')
        x_rolled = roll(x)
        u, s, vt = svd(x_rolled, full_matrices=False)
        uvt_rolled = dot(u, vt)
        # Roll the matrix axes back to their correct positions
        uvt = unroll(uvt_rolled)
        g = expand(g)
        return g * uvt
    else:
        # see https://en.wikipedia.org/wiki/Norm_(mathematics)#p-norm
        return expand(g / ans**(ord-1)) * x * anp.abs(x)**(ord-2)
norm.defvjp(grad_norm)

def grad_eigh(g, ans, vs, gvs, x, UPLO='L'):
    """Gradient for eigenvalues and vectors of a symmetric matrix."""
    N = x.shape[-1]
    w, v = ans              # Eigenvalues, eigenvectors.
    dot = anp.dot if x.ndim == 2 else partial(anp.einsum, '...ij,...jk->...ik')
    wg, vg = g          # Gradient w.r.t. eigenvalues, eigenvectors.
    w_repeated = anp.repeat(w[..., anp.newaxis], N, axis=-1)
    off_diag = anp.ones((N, N)) - anp.eye(N)
    F = off_diag / (T(w_repeated) - w_repeated + anp.eye(N))
    return dot(v * wg[..., anp.newaxis, :] + dot(v, F * dot(T(v), vg)), T(v))
eigh.defvjp(grad_eigh)

def grad_cholesky(g, L, vs, gvs, A):
    # Based on Iain Murray's note http://arxiv.org/abs/1602.07527
    # scipy's dtrtrs wrapper, solve_triangular, doesn't broadcast along leading
    # dimensions, so when A.ndim > 2 we just call a generic LU solve instead of
    # directly using backsubstitution (also, we factor twice...)
    from ..scipy.linalg import solve_triangular
    if anp.ndim(A) == 2:
        solve_trans = partial(solve_triangular, lower=True, trans='T')
    else:
        solve_trans = lambda a, b: solve(T(a), b)
    phi = lambda X: anp.tril(X) / (1. + anp.eye(X.shape[-1]))
    def conjugate_solve(L, X):
        'X -> L^{-T} X L^{-1}'
        return solve_trans(L, T(solve_trans(L, T(X))))

    S = conjugate_solve(L, phi(anp.einsum('...ki,...kj->...ij', L, g)))
    return (S + T(S)) / 2.
cholesky.defvjp(grad_cholesky)

def grad_svd(g, usv, vs, gvs, a, full_matrices=True, compute_uv=True):
    dot = anp.dot if a.ndim == 2 else partial(anp.einsum, '...ij,...jk->...ik')

    if not compute_uv:
        s = usv

        # Need U and V so do the whole svd anyway...
        usv = svd(a, full_matrices=False)
        u = usv[0]
        v = T(usv[2])

        return dot(u * g[..., anp.newaxis, :], T(v))

    elif full_matrices:
        raise NotImplementedError(
            "Gradient of svd not implemented for full_matrices=True")

    else:
        u = usv[0]
        s = usv[1]
        v = T(usv[2])

        m, n = a.shape[-2:]

        k = anp.min((m, n))
        # broadcastable identity array with shape (1, 1, ..., 1, k, k)
        i = anp.reshape(anp.eye(k), anp.concatenate((anp.ones(a.ndim - 2, dtype=int), (k, k))))

        f = 1 / (s[..., anp.newaxis, :]**2 - s[..., :, anp.newaxis]**2 + i)

        if m < n:
            gu = g[0]
            gs = g[1]
            gv = T(g[2])

            utgu = dot(T(u), gu)
            vtgv = dot(T(v), gv)

            i_minus_vvt = (anp.reshape(anp.eye(n), anp.concatenate((anp.ones(a.ndim - 2, dtype=int), (n, n)))) -
                            dot(v, T(v)))

            t1 = (f * (utgu - T(utgu))) * s[..., anp.newaxis, :]
            t1 = t1 + i * gs[..., :, anp.newaxis]
            t1 = t1 + s[..., :, anp.newaxis] * (f * (vtgv - T(vtgv)))

            t1 = dot(dot(u, t1), T(v))

            t1 = t1 + dot(dot(u / s[..., anp.newaxis, :], T(gv)), i_minus_vvt)

            return t1

        elif m == n:
            gu = g[0]
            gs = g[1]
            gv = T(g[2])

            utgu = dot(T(u), gu)
            vtgv = dot(T(v), gv)

            t1 = (f * (utgu - T(utgu))) * s[..., anp.newaxis, :]
            t1 = t1 + i * gs[..., :, anp.newaxis]
            t1 = t1 + s[..., :, anp.newaxis] * (f * (vtgv - T(vtgv)))

            t1 = dot(dot(u, t1), T(v))

            return t1

        elif m > n:
            gu = g[0]
            gs = g[1]
            gv = T(g[2])

            utgu = dot(T(u), gu)
            vtgv = dot(T(v), gv)

            i_minus_uut = (anp.reshape(anp.eye(m), anp.concatenate((anp.ones(a.ndim - 2, dtype=int), (m, m)))) -
                            dot(u, T(u)))

            t1 = (f * (utgu - T(utgu))) * s[..., anp.newaxis, :]
            t1 = t1 + i * gs[..., :, anp.newaxis]
            t1 = t1 + s[..., :, anp.newaxis] * (f * (vtgv - T(vtgv)))

            t1 = dot(dot(u, t1), T(v))

            t1 = t1 + dot(i_minus_uut, dot(gu, T(v) / s[..., :, anp.newaxis]))

            return t1
svd.defvjp(grad_svd)
